!+
! Module: evlistcomb_mod
!
! Authors:  J. Ballet, M. Lortholary - SAp  CEA Saclay
!          
! Description: Merge event lists from all CCD/nodes into a single one
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module evlistcomb_mod

  use emutils_mod
  use eventAttributes
  use datafilter  
  use dal
  use param  
  use errorHandling

  implicit none
  
  private 

! Number of modes treated in parallel

    integer, parameter :: nummodes = 2

! Names of the task parameters 
! imagingset and timingset should not be modified

    character(len=*), parameter :: inevent  = "eventsets", &
                                   instrume = "instrument", &
                                   colnames = "colnames", &
                                   coltypes = "coltypes", &
                                   maintabl = "maintable", &
                                   minimli  = "minattributes", &
                                   maximli  = "maxattributes", &
                                   attribli = "mainattributes", &
                                   tablesli = "othertables", &
                                   prcompat = "primarychecks", &
                                   pronly   = "primaryonly", &
                                   excompat = "extensionchecks"

    character(len=3), dimension(nummodes), parameter :: &
       shortmode = (/"img","tim"/)
    character(len=7), dimension(nummodes), parameter :: &
       longmode = (/"imaging","timing "/)

! Name of the column for CCD/node

    character(len=*), parameter :: ccd_colname = "CCDNR"

! Global variables for receiving parameters values
! Note: 8 char for attributes (keywords)
!      80 char for tables (extension) and columns
! maintabname: Name of the main extension (to merge)
! tablelist  : Names of the secondary extensions (to propagate)
! primcompa  : Names of the primary keywords to check for compatibility
! primonly   : Names of the primary keywords which must not be propagated
! extcompa   : Names of the extension keywords to check for compatibility
! attrilist  : Names of the other extension keywords to propagate
! minlist    : Names of keywords of the main extension to minimise
! maxlist    : Names of keywords of the main extension to maximise
! numlistyp  : Number of columns in Imaging and Timing modes
! datamode   : Values of the DATAMODE keyword in Imaging and Timing modes

    character(len=80),allocatable, dimension(:) :: maintabname
    character(len=80),allocatable, dimension(:) :: tablelist
    character(len=8), allocatable, dimension(:) :: attrilist,minlist,maxlist, &
                                                   primcompa,primonly,extcompa
    integer,          dimension(nummodes)       :: numlistyp
    character(len=80), dimension(nummodes)      :: datamode

! Names, types of the columns for modes Imaging and Timing
    type colDef
       character(len=80)     name
       integer               type
    end type colDef
    type(colDef), allocatable, dimension(:,:)    :: columnDef

  public evlistcomb

contains 

  !---------------------------------------------------------------------
  subroutine evlistcomb
  !---------------------------------------------------------------------

    character(len=1024) :: outevent
    character(len=1024), dimension(nummodes) :: out_name
    integer :: numevfiles, dmode, nfile
    logical, dimension(nummodes) :: onmode
    character(len=1024), allocatable, dimension(:) :: file_names

! Read parameters

    call getparamvalues()

! Read the list of input event files and parse it

    numevfiles = parameterCount(inevent)
    write(errstr,*) numevfiles, " input event files"
    call message(AppMsg,Sparse)
    allocate(file_names(numevfiles))
    do nfile=1,numevfiles
       file_names(nfile) = stringParameter(inevent,nfile-1)
    enddo

! Read the output file names

    do dmode = 1, nummodes
       outevent = trim(longmode(dmode)) // "set"
       out_name(dmode) = stringParameter(trim(outevent))
       if (len_trim(out_name(dmode)) == 0) then
          onmode(dmode) = .false.
          write(errstr,*) "No output required for mode ", &
                          trim(datamode(dmode))
          call message(AppMsg,Sparse)
       else
          onmode(dmode) = .true.
       endif
    enddo

! Sort the event files into modes and merge them

    call getGoodFiles(file_names,maintabname,numevfiles)
    if (numevfiles < 1)  call error("main02","No valid input event file")
    do dmode = 1, nummodes
       write(errstr,*) "Entering mode ", trim(datamode(dmode))
       call message(AppMsg,Sparse)
       if (.not. onmode(dmode))  cycle
       call sortFiles(file_names,numevfiles,maintabname,dmode,out_name(dmode))
    enddo

    deallocate(file_names)
    deallocate(attrilist, minlist, maxlist, tablelist, primcompa, extcompa,  &
               primonly, columnDef, maintabname)

  end subroutine evlistcomb

! Enter input files into the DAL. Test existence of EVENTS extension.

  subroutine getGoodFiles(file_names,ext_name,numfiles)

! file_names : array of file names
! ext_name   : name of extensions to test
! numfiles   : number of files to consider (modified in output)

    character(len=*), dimension(:), intent(inout) :: file_names
    integer, intent(inout) :: numfiles
    character(len=*), dimension(:), intent(in)    :: ext_name

    type(DataSetT) :: in_set, ref_set
    integer :: nfile, goodfiles, nexten, k
    logical :: ok

    goodfiles = 0
    nexten = size(ext_name)
    do nfile = 1, numfiles
       if (setExists(file_names(nfile))) then
          goodfiles = goodfiles + 1

! At this stage open all files in low memory mode (header only)

          in_set = dataSet(file_names(nfile), Read, LOW_MEMORY)
          ok = .true.
          do k = 1,nexten
             if (.not. hasBlock(in_set, ext_name(k))) then
                write(errstr,*) "extension ", trim(ext_name(k)),  &
                     " not found in ", trim(name(in_set)), ". File not merged."
                call warning("getGoodFiles10")
                ok = .false.
             endif
          enddo
          if (.not. ok) then
             call release(in_set)
             goodfiles = goodfiles - 1
             cycle
          endif
          if (goodfiles == 1) then
             ref_set = in_set
          else if (.not. equalKeywords(attributable(ref_set),  &
                                       attributable(in_set),primcompa)) then
             write(errstr,*) trim(name(in_set)), " is rejected."
             call warning("getGoodFiles12")
             call release(in_set)
             goodfiles = goodfiles - 1
             cycle
          else
             call release(in_set)
          endif

! Everything OK. Use first file as reference for comparison with next ones
! Put file name among good ones

          file_names(goodfiles) = file_names(nfile)
       else
          write(errstr,*) trim(file_names(nfile)), " does not exist"
          call warning("getGoodFiles11")
       endif
    enddo
    if (goodfiles > 0)  call release(ref_set)
    numfiles = goodfiles

  end subroutine getGoodFiles

! Sort good files from bad, Timing from Imaging
! The mode is identified by the keywords DATAMODE or DATATYPE

  subroutine sortFiles(file_names, numfiles, ext_name, dmode, out_name)

! file_names : array of file names
! numfiles   : number of files to consider
! ext_name   : name of extensions to merge
! dmode      : mode as defined in datamode (0 = indifferent)
! out_name   : name of output event file

    character(len=*), dimension(:), intent(in) :: file_names, ext_name
    integer, intent(in) :: numfiles, dmode
    character(len=*), intent(in) :: out_name

    type(DataSetT) :: out_set, in_set
    type(TableT)   :: in_tab
    type(ColumnT)  :: out_col
    type(TableT), dimension(:), allocatable :: out_tab
    integer,      dimension(numfiles)       :: num_img
    integer,      dimension(:), allocatable :: numrows
    integer :: nfile, nrows, ccdmax, icol, ikey, numimag, nexten, k
    logical :: badkeywords, createdOutput
    character(len=16)   :: instname, name_col
    character(len=80)   :: shortname, un1, un2, obsSubMode

    if (dmode > nummodes)  then
       write(errstr,*) "called with invalid mode =", dmode,  &
                       ". Contact developer"
       call error("sortFiles01")
    endif

    numimag = 0
    num_img = 0
    nexten = size(ext_name)
    createdOutput = .false.
    obsSubMode = 'Unknown'     ! Default value
fs: do nfile = 1, numfiles

! At this stage open all files in low memory mode (header only)

       in_set = dataSet(file_names(nfile), Read, LOW_MEMORY)
       badkeywords = .false.

       do k = 1, nexten
          in_tab = table(in_set, ext_name(k))

! Test mode (first extension to merge only)

          if (dmode > 0 .and. k == 1) then
             call testMode(in_tab,dmode,badkeywords,obsSubMode)
             if (badkeywords) then   ! Do not send a warning here
                call release(in_set)
                cycle fs
             endif
          endif

! Test whether one may retrieve CCDNR

          if (.not. hasColumn(in_tab, ccd_colname))  &
               badkeywords = (getCcd(block(in_tab)) <= 0)

! Test the coherence of the file's keywords

          if (.not.badkeywords .and. numimag == 0) then

! Open the output event file

             if ( k == 1) then
                createdOutput = .true.
                write(errstr,*) "Create ", trim(out_name)
                call message(AppMsg,Sparse)
                out_set = dataSet(out_name, Create)

! Propagate primary keywords taken from the first file

                write(errstr,*) "Copy general attributes"
                call message(AppMsg,Verbose)
                call setAttributes(attributable(out_set),attributable(in_set))
                allocate(out_tab(nexten), numrows(nexten))
                numrows = 0
             endif

             out_tab(k) = addTable(out_set, ext_name(k), 0, label(in_tab))

          else if (.not. badkeywords) then

! Test the coherence of the file's keywords with the first file's

             badkeywords = .not. equalKeywords(attributable(in_tab),  &
                                 attributable(out_tab(k)), extcompa, .false.)
             call testColKeywords(in_tab, out_tab(k), dmode, badkeywords)
          endif

! Do not go on if file was not accepted

          if (badkeywords) then
             write(errstr,*) trim(name(in_set)), " is rejected."
             call warning("sortFiles12")
             call release(in_set)
             cycle fs
          endif
       enddo

! Second loop over extensions after all checks

       do k = 1, nexten
          in_tab = table(in_set, ext_name(k))

! This creates the columns when necessary

          call updateColKeywords(in_tab, out_tab(k), dmode)

! Propagate a few select keywords to the output table

          do ikey = 1, size(minlist)
             call updateMinMax(attributable(in_tab),attributable(out_tab(k)), &
                               minlist(ikey), 1)
          enddo
          do ikey = 1, size(maxlist)
             call updateMinMax(attributable(in_tab),attributable(out_tab(k)), &
                               maxlist(ikey), 2)
          enddo
          call putSelectKeywords(in_tab, out_tab(k), extcompa)
          call putSelectKeywords(in_tab, out_tab(k), attrilist)
          call copyEventAttributes(in_tab, out_tab(k))

          nrows = numberOfRows(in_tab)
          numrows(k) = numrows(k) + nrows
          write(errstr,*) trim(qualifiedName(in_tab))," has", nrows," rows"
          call message(AppMsg,Verbose)
       enddo
       numimag = numimag + 1
       num_img(numimag) = nfile
       call addAttributes(attributable(out_set), attributable(in_set))
       call release(in_set)
    enddo fs

    if (createdOutput) then

! Write the DATAMODE and SUBMODE keywords

       if (dmode > 0) then
          if (.not. hasAttribute(out_set,"DATAMODE"))  &
             call setAttribute(out_set, "DATAMODE", trim(datamode(dmode)),  &
                               "Instrument mode (IMAGING or TIMING)")
       else 
          if (hasAttribute(out_set,"DATAMODE"))  &
             call deleteAttribute(out_set,"DATAMODE")
       endif
       call setAttribute(out_set,"SUBMODE",obsSubMode,"Guest Observer mode")

! Remove FILENAME keyword

       if (hasAttribute(out_set,"FILENAME"))  &
             call deleteAttribute(out_set,"FILENAME")

       do k = 1, nexten

! Regenerate the TELAPSE keyword

          if (hasAttribute(out_tab(k),"TSTART") .and.  &
              hasAttribute(out_tab(k),"TSTOP")) then
             un1 = units(out_tab(k),"TSTOP")
             un2 = units(out_tab(k),"TSTART")
             if (len_trim(un1) == 0)  un1 = un2
             if (len_trim(un2) == 0)  un2 = un1
             if (un1 == un2) then
                call setAttribute(out_tab(k), "TELAPSE",  &
                     real64Attribute(out_tab(k),"TSTOP") -  &
                     real64Attribute(out_tab(k),"TSTART"), &
                     trim(un2), comment="Full time interval for the exposure")
             else
                write(errstr,*) "TSTART and TSTOP do not have the same",  &
                     " units ", trim(un1), " vs ", trim(un2), " in ",  &
                     trim(qualifiedName(out_tab(k))),". TELAPSE not generated."
                call warning("sortFiles13")
             endif
          endif

! Test whether the columns were actually created
! Downgraded to a message because the list is common to all extensions to merge

          do icol=1,numlistyp(dmode) 
             name_col = columnDef(icol,dmode)%name
             if (.not. hasColumn(out_tab(k), trim(name_col))) then
                write(errstr,*) "Column ", trim(name_col), " was not found",  &
                     " in any of the input extensions ", trim(ext_name(k)),  &
                     ". Not created in output."
                call message(AppMsg,Verbose)
             endif
          enddo

! Create the CCDNR column

          write(errstr,*) "Create column ", trim(ccd_colname)
          call message(AppMsg,Verbose)
          ccdmax = 19
          shortname = ""
          if (hasAttribute(out_set,"INSTRUME")) then
             instname = stringAttribute(out_set,"INSTRUME")
             select case(trim(instname))
             case ("RGS1", "RGS2")
                ccdmax = 14
                shortname = "CCD number (1/14)"
             case ("EMOS1", "EMOS2")
                ccdmax = 17
                shortname = "1st digit=node (0/1), 2nd digit=CCD (1/7)"
             case ("EPN")
                ccdmax = 12
                shortname = "CCD number (1/12)"
             case default
             end select
          endif
          out_col = addColumn(out_tab(k), ccd_colname, Integer8, "",  &
                              comment=trim(shortname))
          call setAttribute(out_col, "TLMIN", 1,  &
                            comment="Minimum for "//ccd_colname)
          call setAttribute(out_col, "TLMAX", ccdmax,  &
                            comment="Maximum for "//ccd_colname)
          call release(out_col)

! Propagate the global keywords

          call addAttributes(attributable(out_tab(k)), attributable(out_set))
          call delSelectKeywords(attributable(out_tab(k)), primonly)

! Set output table at its final size

          if (numrows(k) > 0) then
             write(errstr,*) trim(qualifiedName(out_tab(k))), " has",  &
                             numrows(k), " rows"
             call message(AppMsg,Sparse)
             call insertRows(out_tab(k), 0, numrows(k))
          else
             write(errstr,*) trim(qualifiedName(out_tab(k))), " is empty"
             call warning("sortFiles14")
          endif
       enddo

! Merge event lists and extensions

       call mergeEvents(file_names, num_img, numimag, out_tab, ext_name)
       deallocate(out_tab,numrows)
    else
       write(errstr,*) "No output for mode ", trim(datamode(dmode))
       call message(AppMsg,Sparse)
    endif

  end subroutine sortFiles

! Test the column-related keywords for coherence

  subroutine testColKeywords(in_tab, out_tab, dmode, badkeywords)

! in_tab      : handle to the  input binary table
! out_tab     : handle to the output binary table
! dmode       : mode as defined in datamode
! badkeywords : true in output if something is wrong

    type(TableT), intent(in)    :: in_tab, out_tab
    integer,      intent(in)    :: dmode
    logical,      intent(inout) :: badkeywords

    type(ColumnT)  :: in_col, out_col
    integer :: icol
    character(len=16) :: name_col

    if (badkeywords)  return

! Test the column-related keywords
! Should do that automatically for ALL column-related keywords ?

    do icol=1,numlistyp(dmode) 
       name_col = columnDef(icol,dmode)%name
       if (.not. (hasColumn(out_tab, trim(name_col)) .and.  &
                  hasColumn( in_tab, trim(name_col)))) cycle

! Column unit and label

        in_col = column( in_tab, trim(name_col), Read)
!       out_col = column(out_tab, trim(name_col), Read)
       out_col = column(out_tab, trim(name_col), Modify)   ! DAL bug
       if (len_trim(units( in_col)) > 0 .and.  &
           len_trim(units(out_col)) > 0) then
          badkeywords = trim(units(in_col)) /= trim(units(out_col))
       endif
       if (badkeywords) then
          write(errstr,*) "input column ", trim(name(in_col)),  &
             " does not have the same units ", trim(units(in_col)),  &
             " as in the previous files ", trim(units(out_col))
          call warning("testColKeywords12")
       endif

! TNULL keyword (only for integer columns)

       if (.not. badkeywords) then
          select case(dataType(out_col))
          case (Integer8, Integer16, Integer32)
             if (nullDefined(out_col)) then
                select case(dataType(in_col))
                case (Integer8, Integer16, Integer32)
                   if (nullDefined(in_col)) then
                      badKeywords = intNullValue(in_col)/=intNullValue(out_col)
                      if (badKeywords) then
                         write(errstr,*) "input column ", trim(name(in_col)), &
                            " does not have the same null value ",  &
                            intNullValue(in_col),  &
                            " as in the previous files ", intNullValue(out_col)
                         call warning("testColKeywords14")
                      endif
                   endif
                case default
                end select
             endif
          case default
          end select
       endif

! WCS keywords (uses the generic routine equalKeywords in emutils_mod)

       if (.not. badkeywords)  &
          badkeywords = .not. equalKeywords(attributable(in_col),  &
                                 attributable(out_col),extcompa,.false.)

       call release( in_col)
       call release(out_col)
       if (badkeywords)  exit
    enddo

    if (badkeywords) then
       write(errstr,*) "File ", trim(qualifiedName(in_tab)), " rejected"
       call warning("testColKeywords13")
       call release(in_tab)
       return
    endif

  end subroutine testColKeywords

! Update the column-related keywords, create the columns

  subroutine updateColKeywords(in_tab, out_tab, dmode)

! in_tab      : handle to the  input binary table
! out_tab     : handle to the output binary table
! dmode       : mode as defined in datamode

    type(TableT),   intent(in)  :: in_tab, out_tab
    integer,        intent(in)  :: dmode

    type(ColumnT)  :: in_col, out_col
    integer :: icol, ikey
    character(len=16) :: name_col

! Loop over the columns

    do icol=1,numlistyp(dmode) 
       name_col = columnDef(icol,dmode)%name
       if (.not. hasColumn(in_tab, trim(name_col))) cycle
       in_col = column( in_tab, trim(name_col), Read)

! Create the column if need be

       if (.not. hasColumn(out_tab, trim(name_col))) then
          out_col=addColumn(out_tab,name_col,columnDef(icol,dmode)%type,  &
             units(in_col),comment=label(in_col),dimensions=dimensions(in_col))
       else

! Update the column label and units

          out_col = column(out_tab, trim(name_col), Modify)
          if (len_trim(label(in_col)) > len_trim(label(out_col)))  &
             call relabel(out_col,label(in_col))
          if (len_trim(units(in_col)) /= 0)  &
             call setUnits(out_col,units(in_col))
       endif

! MIN/MAX keywords

       do ikey = 1, size(minlist)
          call updateMinMax(attributable(in_col), attributable(out_col),  &
                            minlist(ikey), 1)
       enddo
       do ikey = 1, size(maxlist)
          call updateMinMax(attributable(in_col), attributable(out_col),  &
                            maxlist(ikey), 2)
       enddo

! TNULL keyword

       select case(dataType(out_col))
       case (Integer8, Integer16, Integer32)
          select case(dataType(in_col))
          case (Integer8, Integer16, Integer32)
                if (nullDefined(in_col))  &
                   call setNullValue(out_col,intNullValue(in_col))
          case default
          end select
       case default
       end select

! WCS keywords

       do ikey = 1, size(extcompa)
          if (hasAttribute( in_col,extcompa(ikey)))  &
             call setAttribute(out_col,attribute(in_col,extcompa(ikey)))
       enddo

       call release( in_col)
       call release(out_col)
    enddo

  end subroutine updateColKeywords

! Update MIN/MAX keywords for columns

  subroutine updateMinMax(in_col, out_col, keyname, dir)

! in_col      : handle to the  input object (Attributable)
! out_col     : handle to the output object (Attributable)
! keyname     : keyword name
! dir         : direction (1: minimise, 2: maximise)

    type(AttributableT), intent(in) :: in_col, out_col
    character(len=*),    intent(in) :: keyname
    integer,             intent(in) :: dir

    real(kind=double) :: keyval

    if (hasAttribute( in_col,keyname)) then
       if (hasAttribute(out_col,keyname)) then
          select case(dir)
          case (1)
             keyval = min(real64Attribute(out_col,keyname),  &
                          real64Attribute( in_col, keyname))
          case (2)
             keyval = max(real64Attribute(out_col,keyname),  &
                          real64Attribute( in_col, keyname))
          case default
             write(errstr,*) "Argument dir =", dir,  &
                             " should be 1 or 2. Contact developer."
             call warning("updateMinMax11")
          end select
          if (dataType(attribute( in_col,keyname)) == Integer_attribute .and. &
              dataType(attribute(out_col,keyname)) == Integer_attribute) then
             call setAttribute(out_col, keyname, nint(keyval),  &
                               units(out_col,keyname), label(out_col,keyname))
          else
             call setAttribute(out_col, keyname, keyval,  &
                               units(out_col,keyname), label(out_col,keyname))
          endif
       else
          call setAttribute(out_col,attribute(in_col,keyname))
       endif
    endif

  end subroutine updateMinMax

! Test the data mode/type for coherence

  subroutine testMode(in_tab, dmode, badkeywords, obsSubMode)

! in_tab      : handle to the binary table extension
! dmode       : mode as defined in datamode
! badkeywords : true in output if something is wrong
! obsSubMode  : Observing mode as defined by the Guest Observer

    type(TableT),     intent(in)    :: in_tab
    integer,          intent(in)    :: dmode
    logical,          intent(inout) :: badkeywords
    character(len=*), intent(out)   :: obsSubMode

    type(DataSetT) :: in_set
    integer :: im
    character(len=80) :: key_val, key_mode
    logical :: imageMode, timeMode, rimagMode, ctimeMode

    if (badkeywords)  return

! Recognize the data mode (case is ignored)

    key_mode = changeCase(datamode(dmode),2)
    badkeywords = .false.
    in_set = parent(in_tab)
    if (hasAttribute(in_set,"DATAMODE")) then
       key_val = changeCase(stringAttribute(in_set,"DATAMODE"),2)
       if (trim(key_val) /= trim(key_mode)) then
          badkeywords = .true.
          do im = 1, nummodes
             badkeywords = badkeywords .and.  &
                           trim(key_val) /= trim(changeCase(datamode(im),2))
          enddo
          if (badkeywords) then
             write(errstr,*) "data mode = ", trim(key_val),  &
                             " in ", trim(name(in_set)), " is not recognized"
             call warning("testMode11")
          endif
          badkeywords = .true.
       endif
    else
       call getMode(in_tab, imageMode, timeMode, rimagMode, ctimeMode)
       select case (trim(key_mode))
       case ("IMAGING", "SPECTROSCOPY")
          badkeywords = .not. (imageMode .or. rimagMode)
       case ("TIMING", "HIGH TIME RESOLUTION")
          badkeywords = .not. (timeMode .or. ctimeMode)
       case default
          badkeywords = .true.
       end select
    endif

! Get the Guest Observer mode from the central CCD

    if (hasAttribute(in_set,"SUBMODE"))  &
       obsSubMode = stringAttribute(in_set,"SUBMODE")

    if (badkeywords)  call release(in_tab)

  end subroutine testMode

! Merge event list files of a single type (Imaging or Timing)

  subroutine mergeEvents(file_names,num_img,numimag,out_tab,ext_name)

! file_names : array of file names
! num_img    : table of selected files among in_set
! numimag    : number of selected files
! out_tab    : handle to the output event table

    character(len=*), dimension(:), intent(in) :: file_names, ext_name
    type(TableT),     dimension(:), intent(in) :: out_tab
    integer,          dimension(:), intent(in) :: num_img
    integer,                        intent(in) :: numimag

! Declarations for the DAL

    type(DataSetT) :: out_set, in_set
    type(TableT)   :: in_tab

    integer, dimension(:), allocatable :: cur_row
    integer :: ncur, nrows, nfile, next, nexten
    integer :: icol, enddir
    character(len=1024) :: fullname
    character(len=80)   :: shortname

    if (numimag > size(num_img,1)) then
       write(errstr,*) "numimag =", numimag, &
           " larger than dimension of num_img =", size(num_img,1),  &
           ". Contact developer"
       call error("mergeEvents01")
    else if (numimag < 1) then
       write(errstr,*) "Number of files selected =",numimag,". Nothing done."
       call warning("mergeEvents10")
       return
    endif

    if (maxval(num_img(1:numimag)) > size(file_names)) then
       write(errstr,*) "max(num_img) =", maxval(num_img(1:numimag)), &
           " larger than dimension of file_names =", size(file_names,1),  &
           ". Contact developer"
       call error("mergeEvents02")
    else if (minval(num_img(1:numimag)) < 1) then
       write(errstr,*) "min(num_img) =", &
           minval(num_img(1:numimag)), " negative", ". Contact developer"
       call error("mergeEvents03")
    endif

! Second loop on input event files to copy over the event lists

!    out_set = parent(out_tab)
    nexten = size(ext_name)
    allocate(cur_row(nexten))
    cur_row = 0
    do ncur = 1, numimag
       nfile = num_img(ncur)
       fullname = file_names(nfile)
       write(errstr,*) "Copying ",trim(fullname)
       call message(AppMsg,Verbose)
       enddir = index(fullname,'/',.true.)
       shortname = "Add file " // fullname(enddir+1:len_trim(fullname))
       in_set = dataSet(file_names(nfile), Read)

! Loop over extensions to merge

       do next = 1, nexten
          call addHistory(out_tab(next),shortname)
          in_tab = table(in_set, ext_name(next))

! Propagate the data (CCDNR is treated separately)

          nrows = numberOfRows(in_tab)
          if (nrows > 0) then
             call seek(out_tab(next), cur_row(next), nrows)
             do icol = 0, numberOfColumns(out_tab(next))-2
                call copyData(in_tab, out_tab(next), icol, nrows)
             enddo

! IF CCDNR already exists, propagate it normally.
! Otherwise make it up from the keywords

             if (hasColumn(in_tab, ccd_colname)) then
                icol = numberOfColumns(out_tab(next)) - 1
                call copyData(in_tab, out_tab(next), icol, nrows)
             else
                call copyCcd(in_tab, out_tab(next))
             endif
          endif
          call release(in_tab)
          cur_row(next) = cur_row(next) + nrows
       enddo

! Loop to copy the exposure and badpix extensions

!       do next = 1, size(tablelist)
!          call copyCcdExtension(in_set, tablelist(next), out_set)
!       enddo

       call release(in_set)
    enddo
    do next = 1, nexten
       call release(out_tab(next))
    enddo
    deallocate(cur_row)

! Third loop on input event files to copy over the extensions
! Somewhat faster in low memory mode but slower in high memory mode

    out_set = parent(out_tab(1))
    do ncur = 1, numimag
       nfile = num_img(ncur)
       in_set = dataSet(file_names(nfile), Read)

! Loop to copy the exposure and badpix extensions

       do next = 1, size(tablelist)
          call copyCcdExtension(in_set, tablelist(next), out_set)
       enddo

       call release(in_set)
    enddo

! Close output event file

    write(errstr,*) "Writing ",trim(name(out_set)), " to disk ..."
    call message(AppMsg,Verbose)
    call release(out_set)

  end subroutine mergeEvents

! Copy select keywords from an extension to another

  subroutine putSelectKeywords(in_tab, out_tab, evKey)

!  in_tab : Handle to the  input table
! out_tab : Handle to the output table
!   evKey : List of keywords

    type(TableT), intent(in) :: in_tab, out_tab
    character(len=*), dimension(:), intent(in) :: evKey
  
    character(len=8) :: keyword
    integer :: ncur

    do ncur = 1, size(evKey)
       keyword = evKey(ncur)
       if (hasAttribute(in_tab,keyword) .and.  &
           .not. hasAttribute(out_tab,keyword))  &
          call setAttribute(out_tab,attribute(in_tab,keyword))
    enddo

  end subroutine putSelectKeywords

! Remove select keywords from an extension

  subroutine delSelectKeywords(in_tab, evKey)

!  in_tab : Handle to the table
!   evKey : List of keywords

    type(AttributableT),            intent(in) :: in_tab
    character(len=*), dimension(:), intent(in) :: evKey
  
    character(len=8) :: keyword
    integer :: ncur

    do ncur = 1, size(evKey)
       keyword = evKey(ncur)
       if (hasAttribute(in_tab,keyword))  call deleteAttribute(in_tab,keyword)
    enddo

  end subroutine delSelectKeywords

! Copy over one column into longer one with possible type change

  subroutine copyData(in_tab, out_tab, icol, nrows)

!  in_tab  : handle to the  input table for the DAL
! out_tab  : handle to the output table for the DAL
! icol     : column number in output
! nrows    : number of rows in input column

    type(TableT), intent(in) :: in_tab, out_tab
    integer, intent(in) :: icol, nrows

    type(ColumnT)  :: in_col, out_col
    integer :: coltype, out_type, irow, nrow1
    character(len=80) :: col_name, expression
    real(kind=double)   :: null1
    integer(kind=int32) :: null0
    real(kind=single),   parameter :: real0 = 0.
    integer(kind=int32), parameter :: long0 = 0
    integer(kind=int16), parameter :: intg0 = 0
    integer(kind=int8),  parameter :: byte0 = 0

! Possible types of columns

    real(kind=double),   dimension(:), pointer   :: dbledata
    real(kind=single),   dimension(:), pointer   :: sngldata
    integer(kind=int32), dimension(:), pointer   :: longdata, fromRow, toRow
    integer(kind=int16), dimension(:), pointer   :: intgdata
    integer(kind=int8),  dimension(:), pointer   :: bytedata
    logical(kind=bool),  dimension(:), pointer   :: lgcldata
    real(kind=double),   dimension(0:nrows-1)    :: buffer
    logical(kind=bool),  dimension(0:nrows-1)    :: buflog

    out_col = column(out_tab, icol, Modify)
    col_name = name(out_col)
    out_type = dataType(out_col)
    null1 = huge(buffer(0))

! Read input column

    if (hasColumn(in_tab, trim(col_name)))  then
       in_col = column(in_tab, col_name, Read)
       coltype = dataType(in_col)
       select case(coltype)
       case (Real64)
          dbledata => real64Data(in_col)
          buffer = dbledata
       case (Real32)
          sngldata => real32Data(in_col)
          buffer = sngldata
       case (Integer32)
          longdata => int32Data(in_col)
          buffer = longdata
       case (Integer16)
          intgdata => int16Data(in_col)
          buffer = intgdata
       case (Integer8)
          bytedata => int8Data(in_col)
          buffer = bytedata
       case (Boolean)
          if (out_type == Boolean) then
             lgcldata => boolData(in_col)
             buflog = lgcldata
          else if (out_type /= String) then
             write(errstr,*) "Transtyping is not allowed for booleans.",  &
                             " Column ", trim(qualifiedName(in_col)),  &
                             " will be set to null in output"
             call warning("copyData11")
             buffer = null1
          endif
       case (String)
          if (out_type == String) then
             do irow = 0, nrows-1
                call setStringCell(out_col,irow,stringCell(in_col,irow))
             enddo
          else if (out_type /= Boolean) then
             write(errstr,*) "Transtyping is not allowed for strings.",  &
                             " Column ", trim(qualifiedName(in_col)),  &
                             " will be set to null in output"
             call warning("copyData12")
             buffer = null1
          endif
       case default
          write(errstr,*) "input column ", trim(qualifiedName(in_col)),  &
             " has an unsupported type ", coltype, ". Contact developer"
          call error("copyData01")
       end select

! Special case for Nulls when transtyping (not Booleans or Strings)

       if (out_type /= coltype) then
          select case(coltype)
          case (Boolean, String)     ! Do nothing in that case
          case default
             write(expression,'("isnull(",A,")")') trim(col_name)
             call initFilter(expression)
             nrow1 = doSelect(in_tab, fromRow, toRow)
             if (nrow1 > 0) then
                do irow = 0,size(fromRow)-1
                   buffer(fromRow(irow):toRow(irow)) = null1
                enddo
             endif
          end select
       endif

       call release(in_col)
    else
       write(errstr,*) "input column ", trim(col_name), " in ",  &
                       trim(qualifiedName(in_tab)), &
                       " does not exist. Set to null in output"
       call warning("copyData10")
       buffer = null1
    endif

! Write column to output
! Test for overflow
! Could use columnSeek instead

    select case(out_type)
    case (Real64)
       dbledata => real64Data(out_col)
       dbledata = buffer
    case (Real32)
       sngldata => real32Data(out_col)
       sngldata = buffer
       if (coltype == Real64) then
          where (abs(buffer) > huge(real0))
             buffer = null1
          endwhere
       endif
    case (Integer32)
       longdata => int32Data(out_col)
       longdata = nint(buffer,int32)
       if (coltype == Real64 .or. coltype == Real32) then
          where (abs(buffer) > huge(long0))
             buffer = null1
          endwhere
       endif
    case (Integer16)
       intgdata => int16Data(out_col)
! nint does not work with kind=int16 on NAGware
       intgdata = int(buffer+sign(0.5D0,buffer),int16)
       if (coltype == Real64 .or. coltype == Real32 .or.  &
           coltype == Integer32) then
          where (abs(buffer) > huge(intg0))
             buffer = null1
          endwhere
       endif
    case (Integer8)
       bytedata => int8Data(out_col)
! nint does not work with kind=int8 on NAGware
       bytedata = int(buffer+sign(0.5D0,buffer),int8)
       if (coltype /= out_type) then
          where (abs(buffer) > huge(byte0))
             buffer = null1
          endwhere
       endif
    case (Boolean)
       if (coltype /= Boolean) then
          write(errstr,*) "Transtyping to booleans is not allowed. Column ",  &
             trim(qualifiedName(in_col))," will be set to False in output"
          call warning("copyData13")
          buflog = .false.
       endif
       lgcldata => boolData(out_col)
       lgcldata = buflog
    case (String)
       if (coltype /= String) then
          write(errstr,*) "Transtyping to strings is not allowed. Column ",  &
             trim(qualifiedName(in_col))," will be set to blank in output"
          call warning("copyData14")
          do irow = 0, nrows-1
             call setStringCell(out_col,irow,'')
          enddo
       endif
    case default
       write(errstr,*) "output column ", trim(qualifiedName(out_col)),  &
          " has an unsupported type ", out_type, ". Contact developer"
       call error("copyData02")
    end select

! Special case for Nulls when transtyping (not Booleans or Strings)
! DAL types are always signed, but Bytes are not in FITS

    select case(out_type)
    case (Boolean, String)
    case default
       if (out_type /= coltype .and. any(buffer==null1)) then
          if (.not. nullDefined(out_col)) then
             select case(out_type)
             case (Integer32)
                null0 = - huge(long0)
              case (Integer16)
                null0 = - huge(intg0)
              case (Integer8)
                null0 = huge(byte0)
             case default
                write(errstr,*) "output column ",  &
                                trim(qualifiedName(out_col)),  &
                                " has an unsupported integer type ",  &
                                out_type, ". Contact developer"
                call error("copyData03")
             end select
             call setNullValue(out_col,null0)
          endif
          do irow = 0,nrows-1
             if ( buffer(irow) == null1 )  call setToNull(out_col,irow)
          enddo
       endif
    end select

    call release(out_col)

  end subroutine copyData

! Write the CCD/node as a column

  subroutine copyCcd(in_tab, out_tab)

!  in_tab  : handle to the  input table for the DAL
! out_tab  : handle to the output table for the DAL

    type(TableT), intent(in) :: in_tab, out_tab

    type(ColumnT)  :: out_col
    integer(kind=int8) :: ccd
    integer(kind=int8),  dimension(:), pointer   :: bytedata

! Read the CCD/node

    ccd = getCcd(block(in_tab))

! Write the CCD/node

    out_col = column(out_tab, ccd_colname, Modify)
    bytedata => int8Data(out_col)
    bytedata = ccd
    call release(out_col)

  end subroutine copyCcd

! Copy extension indexed with CCD/node number

  subroutine copyCcdExtension(in_set, ext_name, out_set)

! in_set     : handle to the input file
! ext_name   : name of extension to copy
! out_set    : handle to the output data set

    type(DataSetT), intent(in) :: in_set
    character(len=*), intent(in) :: ext_name
    type(DataSetT), intent(in) :: out_set

    type(BlockT) :: in_block
    integer :: ccd, ib, ngoodb, nchars
    character(len=8) :: expos_name
    character(len=6) :: ext_trim
    character(len=1024) :: fullname

    nchars = min(6,len_trim(ext_name))
    ext_trim = "      "
    ext_trim(1:nchars) = ext_name(1:nchars)

    fullname = name(in_set)
    if (hasBlock(in_set, ext_name)) then

! Normal case of events file for a single CCD/node

       in_block = block(in_set, ext_name, Read)
       ccd = getCcd(in_block)
       if (ccd > 0) then
          write(expos_name,'(I2.2)') ccd
          expos_name = trim(ext_trim) // expos_name
          call copyExtension(out_set, in_block, expos_name)
       else
          write(errstr,*) "Extension ",trim(ext_name), &
                          " of ",trim(fullname)," not copied."
          call warning("copyCcdExtension10")
       endif
       call release(in_block)
    else

! Look for already indexed extensions

       ngoodb = 0
       do ib = 0, numberOfBlocks(in_set)-1
          in_block = block(in_set,ib,Read)
          expos_name = name(in_block)
          if (expos_name(1:nchars) .eq. ext_name(1:nchars)) then
             call copyExtension(out_set, in_block, expos_name)
             ngoodb = ngoodb + 1
          endif
          call release(in_block)
       enddo
       if (ngoodb == 0) then
          write(errstr,*) "extension ", trim(ext_name), &
                          " not found in ", trim(fullname)
          call message(AppMsg,Sparse)
       endif
    endif

  end subroutine copyCcdExtension

! Copy extension

  subroutine copyExtension(out_set, in_block, expos_name)

! in_block   : handle to the input block
! expos_name : name of extension to copy
! out_set    : handle to the output data set

    character(len=*), intent(in) :: expos_name
    type(DataSetT),   intent(in) :: out_set
    type(BlockT),     intent(in) :: in_block

    type(BlockT) :: out_block

    if (hasBlock(out_set,expos_name)) then
       write(errstr,*) "Extension ", trim(expos_name), " already exists ",  &
          "in output file. Not copied from ", trim(qualifiedName(in_block))
       call warning("copyExtension12")
    else
       write(errstr,*) "Writing extension ", trim(expos_name),  &
                       " from ", trim(qualifiedName(in_block))
       call message(AppMsg,Sparse)
       call copyBlock(out_set,in_block,expos_name)

! Propagate the global keywords

       out_block = block(out_set, expos_name, Modify)
       call addAttributes(attributable(out_block),attributable(out_set))
       call delSelectKeywords(attributable(out_block), primonly)
       call release(out_block)
    endif

  end subroutine copyExtension

! Parameter reading

  subroutine getparamvalues()

! In globals:
! Parameter names: emimlist, emimtype, emtilist, emtitype
!                  rgimlist, rgimtype, rgtilist, rgtitype
!                  epimlist, epimtype, eptilist, eptitype
! Global variables: maintabname, tablelist
!                   attrilist, primcompa, primonly, extcompa, numlistyp

    character(len=5) :: instrumname
    integer  :: listLength

! Read instrume

    instrumname = stringParameter(instrume)
    write(errstr,*) "Instrument name : ",trim(instrumname)
    call message(AppMsg,Sparse)

! Read the column parameters corresponding to the instrument

    call readColParams(trim(instrumname))

! Read maintabl with conversion in uppercase

    listLength = sizeListParam(maintabl)
    allocate( maintabname(listLength) )
    call readListParam(maintabl,maintabname,listLength)
    if (listLength > 0) then
       write(errstr,*) "Main table name : ",trim(maintabname(1))
       call message(AppMsg,Sparse)
    else
       write(errstr,*) "At least one table must be merged. ", trim(maintabl), &
                       " cannot be empty."
       call error("getparamvalues01")
    endif

! Read the lists

    listLength = sizeListParam(attribli)
    allocate( attrilist(listLength) )
    call readListParam(attribli,attrilist,listLength)

    listLength = sizeListParam(minimli)
    allocate( minlist(listLength) )
    call readListParam(minimli,minlist,listLength)

    listLength = sizeListParam(maximli)
    allocate( maxlist(listLength) )
    call readListParam(maximli,maxlist,listLength)

    listLength = sizeListParam(tablesli)
    allocate( tablelist(listLength) )
    call readListParam(tablesli,tablelist,listLength)

    listLength = sizeListParam(prcompat)
    allocate( primcompa(listLength) )
    call readListParam(prcompat,primcompa,listLength)

    listLength = sizeListParam(pronly)
    allocate( primonly(listLength) )
    call readListParam(pronly,primonly,listLength)

    listLength = sizeListParam(excompat)
    allocate( extcompa(listLength) )
    call readListParam(excompat,extcompa,listLength)

  end subroutine getparamvalues

! Read the column characteristics

  subroutine readColParams(instrumname)

! instrumname : Prefix column parameters

    character(len=*), intent(in) :: instrumname

    character(len=80), allocatable, dimension(:) :: collist, typelist
    character(len=80) :: parnames, partypes
    integer :: numcolmax, ncol, dmode, nnlist, nntype

! Read DATAMODE values

    parnames = instrumname // 'datamodes'
    nnlist = sizeListParam(trim(parnames))
    if (nnlist > nummodes) then
       write(errstr,*) "evlistcomb may not distinguish more than", nummodes,  &
          " modes. Further values in ", trim(parnames), " are ignored."
       call warning('readColParams11')
       nnlist = nummodes
    endif
    call readListParam(trim(parnames),datamode,nnlist)

! First loop over modes to get longest column list

    numcolmax = 0
    do dmode = 1,nummodes
       parnames = instrumname // shortmode(dmode) // colnames
       nnlist = sizeListParam(trim(parnames))
       numcolmax = max(numcolmax,nnlist)
       numlistyp(dmode) = nnlist
    enddo

! Allocate the columnDef structure

    allocate( columnDef(numcolmax,nummodes) )
    columnDef%name = ''
    columnDef%type = 0
    allocate( collist(numcolmax),typelist(numcolmax) )

! Second loop over modes to read column lists

    do dmode = 1,nummodes

! Read the column names

       parnames = instrumname // shortmode(dmode) // colnames
       nnlist = numlistyp(dmode)
       call readListParam(trim(parnames),collist,nnlist)

! Read the column types

       partypes = instrumname // shortmode(dmode) // coltypes
       nntype = sizeListParam(trim(partypes))
       call readListParam(trim(partypes),typelist,nntype)
       if (nnlist /= nntype) then
          write(errstr,*) trim(parnames), " and ", trim(partypes),  &
                             " do not have the same size"
          call error("readColParams05")
       endif

       columnDef(1:nnlist,dmode)%name = collist(1:nnlist)
       do ncol = 1,nnlist
          columnDef(ncol,dmode)%type = typeFromName(typelist(ncol))
       enddo
    enddo

    deallocate(collist,typelist)

  end subroutine readColParams
 
! Merely get the 'integer' type from type name (uppercased)

  function typeFromName( typename ) result( type )
   character(len=*)  :: typename    
   integer           :: type

   select case(trim(changeCase(typename,2)))
    case ("DOUBLE")
      type = Real64   
    case ("SINGLE")
      type = Real32        
    case ("INT32")
      type = Integer32
    case ("INT16")
      type = Integer16
    case ("INT8")
      type = Integer8
    case ("BOOLEAN")
      type = Boolean
    case ("STRING")
      type = String
    case default
      write(errstr,*) "Column type ", trim(typename), " was not recognized."
      call error("typeFromName01")
   end select  

  end function typeFromName

end module evlistcomb_mod
